/**
*	project2.cpp
*	CS 6610 - Advanced Computer Graphics I
*	Zach Gildersleeve
*	Oct. 13, 2006
*	
*	Creates a square room with several objects that demonstrate
*	different aspects of texture mapping.  The room walls and floor
*	are textured with multitextured texture objects.  The cone
*	demonstrates different texture environment modes, and the sphere
*	is mapped with a dynamic reflective cube map environment.
*	Additionally, a dragonfly model incorporates other texture elements.
*	Detailed documentation for this project can be found in "REPORT.html"
*/

#include "main.h"
#include "vector.h"

int cur_button = -1;				// mouse handling for main window
int last_x;							// mouse position
int last_y;							// mouse position

GLUI				*glui;			// pointers for all of the glui controls
GLUI_Panel			*lightPanel;
GLUI_Panel			*shadowPanel;
GLUI_Panel			*depthPanel;
GLUI_Spinner		*lightSpin;
GLUI_Spinner		*triangleSize;
GLUI_Rollout		*shadowRollout;	
GLUI_Rollout		*lightRollout;
GLUI_Rollout		*depthRollout;
GLUI_RadioGroup		*shadowRadio;
GLUI_Translation	*lightTransXY;
GLUI_Translation	*lightTransZ;
GLUI_Checkbox		*volumeCheck;

int showVolumeToggle;				
GLfloat projectiveShadowMatrix[4][4];	// matrix to create projective shadows
GLfloat plane[4];					// plane equation for walls

int shadowMode = 0;					// selected shadow mode, 0 == projective, 1 == shadowmap, 2 == shadowvolumes
int main_window;					// the main window name
float eye[3];						// camera position
float lookat[3];					// camera look position
float lightPositionXY[] = {10.0, 10.0};	// the XY light position
float lightPositionZ = -10.0;		// the Z light position

float lightIntensity = 1.0;			// controls intensity of scene light
float triangleSizeMultiplier = 1.0;
float triangleSizeMultiplier2 = 1.0;
int viewX = 189, viewY, viewW, viewH;	// variables to builds the viewport around the GLUI GUI

GLfloat	 lightPos[]		= {5.0, 5.0, 5.0, 1.0};		// position of light
GLfloat  diffuseLight[] = {1.0, 1.0, 1.0, 0.0};		// color of light

GLfloat textureMatrix[16];			// texture matrix for shadowmaps

static unsigned *image;				// pointer to texture data
static int width;					// width of texture data
static int height;					// height of texture data
static int components;				// components of texture data

GLuint floorTexture;				// name of floor texture
GLuint shadowMap;					// name of shadowmap texture

GLfloat tri1[] = {8.0, 8.0, 0.0};	// vectors of the shadow volume triangle
GLfloat tri2[] = {10.0, 3.0, -2.0};
GLfloat tri3[] = {10.0, 3.0, 2.0};

GLfloat tri1extended[3];			// vectors of the shadow volume triangle
GLfloat tri2extended[3];			// extruded to infinity
GLfloat tri3extended[3];

/**
*	Idle function for GLUT callback
*/
void myGlutIdle(void)
{
	if (glutGetWindow() != main_window)
		glutSetWindow(main_window);
	glutPostRedisplay();
}

/**
*	Mouse callback function for GLUT
*
*	@param	button	the current mouse button
*	@param	state	the state of 'button'
*	@param	x		the x coord of the mouse
*	@param	y		the y coord of the mouse
*/
void myGlutMouse(int button, int state, int x, int y)
{
	if (state == GLUT_DOWN)
		cur_button = button;
	else
	{
		if (button == cur_button)
			cur_button = -1;
	}
	
	last_x = x;
	last_y = y;
}

/**
*	Mouse motion callback function for GLUT
*
*	@param	x		the x coord of the mouse
*	@param	y		the y coord of the mouse
*/
void myGlutMotion(int x, int y)
{
	int dx = x-last_x;			// change in mouse position
	int dy = y-last_y;			// change in mouse position
	
	float scale, len, theta;	// controls camera movement
	float neye[3], neye2[3];	// temp eye positions
	float f[3], r[3], u[3];		// more camera coords
	
	switch(cur_button)
	{
		case GLUT_LEFT_BUTTON:	// translate
			f[0] = lookat[0] - eye[0];
			f[1] = lookat[1] - eye[1];
			f[2] = lookat[2] - eye[2];
			u[0] = 0;
			u[1] = 1;
			u[2] = 0;
			
			scale = sqrt(v_getVectorLength(f)) * 0.007;	// scale the change by how far away we are
			
			v_crossProduct(f, u, r);
			v_crossProduct(r, f, u);
			v_normalize(r);
			v_normalize(u);
			
			eye[0] += -r[0]*dx*scale + u[0]*dy*scale;
			eye[1] += -r[1]*dx*scale + u[1]*dy*scale;
			eye[2] += -r[2]*dx*scale + u[2]*dy*scale;
			
			lookat[0] += -r[0]*dx*scale + u[0]*dy*scale;
			lookat[1] += -r[1]*dx*scale + u[1]*dy*scale;
			lookat[2] += -r[2]*dx*scale + u[2]*dy*scale;
			
			break;
			
		case GLUT_MIDDLE_BUTTON:	// zoom
			f[0] = lookat[0] - eye[0];
			f[1] = lookat[1] - eye[1];
			f[2] = lookat[2] - eye[2];
			
			len = v_getVectorLength(f);
			v_normalize(f);
			
			len -= sqrt(len)*dx*0.03;	// scale the change by how far away we are
			
			eye[0] = lookat[0] - len*f[0];
			eye[1] = lookat[1] - len*f[1];
			eye[2] = lookat[2] - len*f[2];
			
			// make sure the eye and lookat points are sufficiently far away
			// push the lookat point forward if it is too close
			if (len < 1)
			{
				lookat[0] = eye[0] + f[0];
				lookat[1] = eye[1] + f[1];
				lookat[2] = eye[2] + f[2];
			}
			break;
			
		case GLUT_RIGHT_BUTTON:	// rotate
			
			neye[0] = eye[0] - lookat[0];
			neye[1] = eye[1] - lookat[1];
			neye[2] = eye[2] - lookat[2];
			
			// first rotate in the x/z plane
			theta = -dx * 0.007;
			neye2[0] = (float)cos(theta)*neye[0] + (float)sin(theta)*neye[2];
			neye2[1] = neye[1];
			neye2[2] =-(float)sin(theta)*neye[0] + (float)cos(theta)*neye[2];
			
			// now rotate vertically
			theta = -dy * 0.007;
			
			f[0] = -neye2[0];
			f[1] = -neye2[1];
			f[2] = -neye2[2];
			u[0] = 0;
			u[1] = 1;
			u[2] = 0;
			v_crossProduct(f, u, r);
			v_crossProduct(r, f, u);
			len = v_getVectorLength(f);
			v_normalize(f);
			v_normalize(u);
			
			neye[0] = len * ((float)cos(theta)*f[0] + (float)sin(theta)*u[0]);
			neye[1] = len * ((float)cos(theta)*f[1] + (float)sin(theta)*u[1]);
			neye[2] = len * ((float)cos(theta)*f[2] + (float)sin(theta)*u[2]);
			
			eye[0] = lookat[0] - neye[0];
			eye[1] = lookat[1] - neye[1];
			eye[2] = lookat[2] - neye[2];
			
			break;
	}
	
	last_x = x;
	last_y = y;
	
	glutPostRedisplay();
}

/**
*	Switches material attributes, controls object
*	color when lighting is enabled.
*
*	@param	mode	current material object selected
*/
void materialMode(int mode)
{
	GLfloat none[]			= {0.0, 0.0, 0.0, 1.0};
	GLfloat gray[]			= {0.6, 0.6, 0.6, 1.0};
	GLfloat gold[]			= {0.79, 0.71, 0.02, 0.5};
	GLfloat white[]			= {1.0, 1.0, 1.0, 1.0};
	GLfloat lightEmission[] = {1.0, 1.0, 1.0, 0.0};
	GLfloat specular[]		= {1.0, 1.0, 1.0, 1.0};
	GLfloat medShiny[]		= {50.0};
	GLfloat highShiny[]		= {150.0};
	
	switch(mode)
	{
		case ROOM_OBJECT:
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, gray);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, none);
			glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, none);
			break;
		case CONE_OBJECT:
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, gray);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, none);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, highShiny);
			glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, none);
			break;
		case SPHERE_OBJECT:
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, gray);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, medShiny);
			glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, none);	
			break;
		case LIGHT_OBJECT:
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, white);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, highShiny);
			glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, lightEmission);
			break;
		case MIRROR_OBJECT:
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, gold);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular);
			glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, highShiny);
			glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, none);
			break;	
	}
}

/**
*	Initializes the OpenGL state, builds a light,
*	and loads all the textures.  Also loads temp
*	textures for cubemap
*/
GLvoid Init(GLvoid)
{
	GLfloat  ambientLight[] = {0.1, 0.1, 0.2, 0.0};		// color of overall ambient light
	
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
	glEnable(GL_LINE_SMOOTH);
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
		
	glEnable(GL_LIGHTING);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambientLight);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuseLight);
	glEnable(GL_LIGHT0);
	
	glClearColor(0.3, 0.3, 0.3, 0.0);
	glClearDepth(1.0);
	glDepthFunc(GL_LEQUAL);
	glClearStencil(0);
	glGenTextures(1, &floorTexture);		// standard texture objects
	glBindTexture(GL_TEXTURE_2D, floorTexture);
		image = read_texture("/Users/zachgildersleeve/Documents/Soc\ F2006/CP\ SC\ 6610\ Computer\ Graphics/project2/project2/plank.rgb", &width, &height, &components);
		gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA8, width, height, GL_RGBA, GL_UNSIGNED_BYTE, image);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexImage2D(GL_TEXTURE_2D, 0, components, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, image);
		free(image);
	glGenTextures(1, &shadowMap);
		glBindTexture(GL_TEXTURE_2D, shadowMap);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, SHADOWMAP_SIZE, SHADOWMAP_SIZE, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, NULL);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE, GL_LUMINANCE);
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		
	// build texture units from objects, used for multitexturing floor with shadowmap	
	glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, floorTexture);
	
	glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, shadowMap);
}

/**
*	Reshape callback function for GLUT
*
*	@param	x	the x window coord - unused
*	@param	y	the y window coord - unused
*/
void myGlutReshape(int	x, int y)
{
	GLUI_Master.get_viewport_area(&viewX, &viewY, &viewW, &viewH);
	glViewport(viewX, viewY, viewW, viewH);
}

/**
*	Draws the cone.
*/
GLvoid drawCone (GLvoid)
{	
	glPushMatrix();
		materialMode(CONE_OBJECT);
		glTranslatef(0.0, -20.0, 0.0);
		glScalef(5.0, 15.0, 5.0);
		glRotatef(-90.0, 1.0, 0.0, 0.0);
		glutSolidCone(1.0, 1.0, 40, 40);
	glPopMatrix();	
}

/**
*	Draws the sphere.
*/
GLvoid drawSphere(GLvoid)
{
	glPushMatrix();
	glScalef(5.0, 5.0, 5.0);
	materialMode(SPHERE_OBJECT);
	glutSolidSphere(1.0, 50, 50);
	glPopMatrix();
}
/**
*	Draws the interlinked torus.
*/
GLvoid drawTori(GLvoid)
{
	glPushMatrix();
	glTranslatef(0.0, 8.0, 0.0);
	glScalef(2.0, 2.0, 2.0);
	materialMode(SPHERE_OBJECT);
	glRotatef(45.0, 0.0, 1.0, 0.0);
	glutSolidTorus(0.5, 1.0, 50, 50);
	glTranslatef(0.0, 1.0, 0.0);
	glRotatef(90.0, 0.0, 1.0, 0.0);
	glutSolidTorus(0.5, 1.0, 50, 50);
	glPopMatrix();
}

/**
*	Draws the room, and textures the floor.
*/
GLvoid DrawRoom(GLvoid)
{
	glPushMatrix();
		glScalef(20.0, 20.0, 20.0);
		
		glActiveTexture(GL_TEXTURE0);
			glEnable(GL_TEXTURE_2D);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			glBindTexture(GL_TEXTURE_2D, floorTexture);
		glActiveTexture(GL_TEXTURE1);
			glEnable(GL_TEXTURE_2D);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		
		materialMode(ROOM_OBJECT);

		glMatrixMode(GL_MODELVIEW);		
		glBegin(GL_QUADS);
			glNormal3d(0, -1, 0);
			glTexCoord2f(1.0, 0.0);		glVertex3f( 1.0,-1.0, 1.0);
			glTexCoord2f(0.0, 0.0);		glVertex3f(-1.0,-1.0, 1.0);
			glTexCoord2f(0.0, 1.0);		glVertex3f(-1.0,-1.0,-1.0);
			glTexCoord2f(1.0, 1.0);		glVertex3f( 1.0,-1.0,-1.0);
		glEnd();
		
		glActiveTexture(GL_TEXTURE0);
			glDisable(GL_TEXTURE_2D);
			
		glActiveTexture(GL_TEXTURE1);
			glBindTexture(GL_TEXTURE_2D, shadowMap);
			glEnable(GL_TEXTURE_2D);
			
		glBegin(GL_QUADS);
			glNormal3d(0, 1, 0);			// Top
			glVertex3f( 1.0, 1.0,-1.0);
			glVertex3f(-1.0, 1.0,-1.0);
			glVertex3f(-1.0, 1.0, 1.0);
			glVertex3f( 1.0, 1.0, 1.0);

			glNormal3d(0, 0, 1);			// Front
			glVertex3f( 1.0, 1.0, 1.0);
			glVertex3f(-1.0, 1.0, 1.0);
			glVertex3f(-1.0,-1.0, 1.0);
			glVertex3f( 1.0,-1.0, 1.0);
		
			glNormal3d(0, 0, -1);			// Back
			glVertex3f( 1.0,-1.0,-1.0);
			glVertex3f(-1.0,-1.0,-1.0);
			glVertex3f(-1.0, 1.0,-1.0);
			glVertex3f( 1.0, 1.0,-1.0);

			glNormal3d(-1, 0, 0);			// Left
			glVertex3f(-1.0, 1.0, 1.0);
			glVertex3f(-1.0, 1.0,-1.0);
			glVertex3f(-1.0,-1.0,-1.0);
			glVertex3f(-1.0,-1.0, 1.0);
			
			glNormal3d(1, 0, 0);			// Right
			glVertex3f( 1.0, 1.0,-1.0);
			glVertex3f( 1.0, 1.0, 1.0);
			glVertex3f( 1.0,-1.0, 1.0);
			glVertex3f( 1.0,-1.0,-1.0);
		glEnd();
		
	glPopMatrix();
}

/**
*	Calculates the distance between the camera position
*	and the camera lookat point.  Used for gluPerspective().
*
*	@return		the distance calculated
*/
GLfloat cameraDistance(GLvoid)
{
	float dx, dy, dz;
	
	dx = eye[0] - lookat[0];
	dy = eye[1] - lookat[1];
	dz = eye[2] - lookat[2];

	return 	sqrt(dx * dx + dy * dy + dz * dz);
}

/**
*	Builds the light geometry (sphere) for the moving light.
*/
GLvoid buildLight(GLint mode)
{
	// constrain the light to keep it inside the room
	if (lightPositionXY[0] > 19.0)
		lightPositionXY[0] = 19.0;
	if (lightPositionXY[0] < -19.0)
		lightPositionXY[0] = -19.0;
	if (lightPositionXY[1] > 19.0)
		lightPositionXY[1] = 19.0;
	if (lightPositionXY[1] < -19.0)
		lightPositionXY[1] = -19.0;	
	if(lightPositionZ > 19.0)
		lightPositionZ = 19.0;
	if(lightPositionZ < -19.0)
		lightPositionZ = -19.0;	
		
	lightPos[0] = lightPositionXY[0];
	lightPos[1] = lightPositionXY[1];
	
	if (mode == 0)
		lightPos[2] = lightPositionZ;
	else
		lightPos[2] = 38 + lightPositionZ;
	
	printf("lightPos = %f %f %f\n", lightPos[0], lightPos[1], lightPos[2]);
	
	glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuseLight);
	glPushMatrix();
		materialMode(LIGHT_OBJECT);
		glTranslatef(lightPos[0], lightPos[1], lightPos[2]);
		glutSolidSphere(0.5, 15, 15);
	glPopMatrix();
	
	glutPostRedisplay();
	
}

/**
*	Builds a matrix for projective shadows, according to the plane
*	described in groundplane[].
*
*	@param	shadowMat	an array to store the built matrix in
*	@param	groundplane	an array that represents the A, B, C, D values in a plane
*/
void shadowMatrix(GLfloat shadowMat[4][4], GLfloat groundplane[4]) 
{ 
	GLfloat dot;	// the dot product between light position and ground plane normal
	dot =	groundplane[0] * lightPos[0] +
			groundplane[1] * lightPos[1] +
			groundplane[2] * lightPos[2] +
			groundplane[3] * lightPos[3]; 
	shadowMat[0][0] = dot - lightPos[0] * groundplane[0]; 
	shadowMat[1][0] = 0.0 - lightPos[0] * groundplane[1]; 
	shadowMat[2][0] = 0.0 - lightPos[0] * groundplane[2]; 
	shadowMat[3][0] = 0.0 - lightPos[0] * groundplane[3]; 
	shadowMat[0][1] = 0.0 - lightPos[1] * groundplane[0]; 
	shadowMat[1][1] = dot - lightPos[1] * groundplane[1]; 
	shadowMat[2][1] = 0.0 - lightPos[1] * groundplane[2]; 
	shadowMat[3][1] = 0.0 - lightPos[1] * groundplane[3]; 
	shadowMat[0][2] = 0.0 - lightPos[2] * groundplane[0]; 
	shadowMat[1][2] = 0.0 - lightPos[2] * groundplane[1]; 
	shadowMat[2][2] = dot - lightPos[2] * groundplane[2]; 
	shadowMat[3][2] = 0.0 - lightPos[2] * groundplane[3]; 
	shadowMat[0][3] = 0.0 - lightPos[3] * groundplane[0]; 
	shadowMat[1][3] = 0.0 - lightPos[3] * groundplane[1]; 
	shadowMat[2][3] = 0.0 - lightPos[3] * groundplane[2]; 
	shadowMat[3][3] = dot - lightPos[3] * groundplane[3]; 
}

/**
*	Sets the plane equation, and stores values in matrix.
*
*	@param	a	the A component of the plane equation
*	@param	b	the B component of the plane equation
*	@param	c	the C component of the plane equation
*	@param	d	the D component of the plane equation
*/
GLvoid setPlane(GLfloat a, GLfloat b, GLfloat c, GLfloat d)
{
	plane[0] = a;
	plane[1] = b;
	plane[2] = c;
	plane[3] = d;
}

/**
*	Draws the scene from the light's POV, and
*	saves the depth map into a texture.
*/
GLvoid makeShadowMap(GLvoid)
{
	GLfloat lightNearPlane = 0.1;
	GLfloat lightFarPlane = 40.0;
	GLfloat lightFOV = 90.0;
	
	GLfloat lightModelviewMatrix[16];
	GLfloat lightProjectionMatrix[16];
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPerspective(lightFOV, 1.0, lightNearPlane, lightFarPlane);
	glGetFloatv(GL_PROJECTION_MATRIX, lightProjectionMatrix);
	
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
		glLoadIdentity();
		gluLookAt(lightPos[0], lightPos[1], lightPos[2], 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
		glGetFloatv(GL_MODELVIEW_MATRIX, lightModelviewMatrix);
		glViewport(0, 0, SHADOWMAP_SIZE, SHADOWMAP_SIZE);

		glClear(GL_DEPTH_BUFFER_BIT);
		
		glShadeModel(GL_FLAT);
		glDisable(GL_LIGHTING);
		glDisable(GL_NORMALIZE);
		glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

		glEnable(GL_POLYGON_OFFSET_FILL);
		glPolygonOffset(1.0, 0.0);
		drawCone();
		drawSphere();
		drawTori();
		DrawRoom();
		glBindTexture(GL_TEXTURE_2D, shadowMap);
		glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, SHADOWMAP_SIZE, SHADOWMAP_SIZE);

		glShadeModel(GL_SMOOTH);
		glEnable(GL_LIGHTING);
		glEnable(GL_NORMALIZE);
		glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

		glDisable(GL_POLYGON_OFFSET_FILL);
	glPopMatrix();
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	// scale and bias the texture matrix, and
	// multiply by the two light matrix, and
	// store the new texture matrix for later
	glMatrixMode(GL_TEXTURE);
	glPushMatrix();
		glLoadIdentity();
		glTranslatef(0.5, 0.5, 0.0);
		glScalef(0.5, 0.5, 1.0);
		glMultMatrixf(lightProjectionMatrix);
		glMultMatrixf(lightModelviewMatrix);
		glGetFloatv(GL_TEXTURE_MATRIX, textureMatrix);
	glPopMatrix();
}
/**
*	Draws the triangle occluder for the shadow volumes.
*	The triangle is scaled using the GLUI spinner.
*/
GLvoid drawTriangleOccluder(GLvoid)
{
	glPushMatrix();
	glScalef(1.0, triangleSizeMultiplier, triangleSizeMultiplier);
	glBegin(GL_TRIANGLES);
		glVertex3fv(tri1);
		glVertex3fv(tri2);
		glVertex3fv(tri3);
	glEnd();
	glPopMatrix();
}

/**
*	Draws the shadow volume geometry by extruding to "infinity"
*	each respective vertex of the occluder triangle.
*/
GLvoid drawShadowVolumeGeometry(GLvoid)
{
	GLfloat infinity = 200.0;	// what we assume to be the furthest distance of the shadow volume
								// it is much larger than the size of the room, to extrude
								// the geometry along pseudo parallel lines
	glDisable(GL_LIGHTING);
	
	if (showVolumeToggle)
	{
		glEnable(GL_BLEND);
		glColor4f(0.0, 1.0, 0.0, 0.5);		
	}
	
	// extrude each vertex to "infinity", and store point
	// in tri*extended[]
	tri1extended[0] = tri1[0] - lightPos[0];
	tri1extended[1] = tri1[1] - lightPos[1];
	tri1extended[2] = tri1[2] - lightPos[2];
	v_normalize(tri1extended);
	tri1extended[0] *= infinity;
	tri1extended[1] *= infinity;
	tri1extended[2] *= infinity;
	
	tri2extended[0] = tri2[0] - lightPos[0];
	tri2extended[1] = tri2[1] - lightPos[1];
	tri2extended[2] = tri2[2] - lightPos[2];
	v_normalize(tri2extended);
	tri2extended[0] *= infinity;
	tri2extended[1] *= infinity;
	tri2extended[2] *= infinity;
	
	tri3extended[0] = tri3[0] - lightPos[0];
	tri3extended[1] = tri3[1] - lightPos[1];
	tri3extended[2] = tri3[2] - lightPos[2];
	v_normalize(tri3extended);
	tri3extended[0] *= infinity;
	tri3extended[1] *= infinity;
	tri3extended[2] *= infinity;
	
	// draw volume geometry
	glPushMatrix();
		glScalef(1.0, triangleSizeMultiplier, triangleSizeMultiplier);

		glBegin(GL_QUADS);
				glVertex3fv(tri1);
				glVertex3fv(tri1extended);
				glVertex3fv(tri2extended);
				glVertex3fv(tri2);
				
				glVertex3fv(tri2);
				glVertex3fv(tri2extended);
				glVertex3fv(tri3extended);
				glVertex3fv(tri3);
				
				glVertex3fv(tri3);
				glVertex3fv(tri3extended);
				glVertex3fv(tri1extended);
				glVertex3fv(tri1);
		glEnd();
		
		glBegin(GL_TRIANGLES);
			glVertex3fv(tri1extended);
			glVertex3fv(tri2extended);
			glVertex3fv(tri3extended);
		glEnd();
		
	glPopMatrix();
	
	drawTriangleOccluder();
	
	if (showVolumeToggle)
	{
		glDisable(GL_BLEND);
	}
	
	glEnable(GL_LIGHTING);
}

/**
*	Using the drawn shadow volume geometry, fills the
*	stencil buffer with the appropriate value based on the
*	zpass shadow volume algorithm.  The shadow volume geometry
*	is drawn twice, incrementing up for front facing polygons,
*	and decrementing for back facing polygons.  The scene
*	is then drawn where there is no shadow.
*/
GLvoid makeShadowVolume(GLvoid)
{
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask(GL_FALSE);
	glClear(GL_STENCIL_BUFFER_BIT);
	glEnable(GL_STENCIL_TEST);
    glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(0.0, 0.0);
	
	glEnable(GL_CULL_FACE);

	glCullFace(GL_FRONT);
	glStencilFunc(GL_ALWAYS, 0, 1);
	glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);

	drawShadowVolumeGeometry();

	glCullFace(GL_BACK);
	glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
	
	drawShadowVolumeGeometry();
	
	glDepthMask(GL_TRUE);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	
	glStencilFunc(GL_EQUAL, 0, 1);
	glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
	
	drawTori();
	drawCone();
	drawSphere();

	glDisable(GL_CULL_FACE);	
	DrawRoom();
	glDisable(GL_POLYGON_OFFSET_FILL);
	
	glDisable(GL_STENCIL_TEST);
}


/**
*	Switches between the three different shadow methods, and ensures
*	that the state is correct.
*/
GLvoid drawShadows(GLint mode)
{
	GLfloat room[6][3] = {	{0.0, -1.0, 0.0},		// the coordinates of the room, expressed in an
							{0.0, 1.0, 0.0},		// A, B, C, D equation
							{0.0, 0.0, 1.0},
							{0.0, 0.0, -1.0},
							{-1.0, 0.0, 0.0},
							{1.0, 0.0, 0.0}};
	GLfloat lowLight[4] = {0.1 * lightIntensity,	// simulates the ambient light when two passes
						   0.1 * lightIntensity,	// are necessary
						   0.1 * lightIntensity,
						   1.0};
	GLfloat sPlane[4] = {1.0f, 0.0f, 0.0f, 0.0f};	// represents the eye plane with which the
	GLfloat tPlane[4] = {0.0f, 1.0f, 0.0f, 0.0f};	// shadow map's coordinates are generated
	GLfloat rPlane[4] = {0.0f, 0.0f, 1.0f, 0.0f};
	GLfloat qPlane[4] = {0.0f, 0.0f, 0.0f, 1.0f};						
							
	switch (mode)
	{
		// projective shadows - draws the room geometry into the stencil buffer
		// and color/depth buffers, then using the constructed projective shadow
		// matrix for each face of the room, draws the shadows on the walls, etc.
		case 0:		
			glEnable(GL_STENCIL_TEST);
			glStencilFunc(GL_ALWAYS, 1, 1);
			glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
			DrawRoom();

			glStencilFunc(GL_EQUAL, 1, 1);
			glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);

			glPushMatrix();
				glEnable(GL_CULL_FACE);
				glCullFace(GL_FRONT);
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

					glDisable(GL_LIGHTING);

					glColor4f(0.0, 0.0, 0.0, 0.4);			
					for (int i = 0; i < 6; i++)
					{
						glPushMatrix();
							setPlane(room[i][0], room[i][1], room[i][2], 19.9);
							shadowMatrix(projectiveShadowMatrix, plane);
							glMultMatrixf((GLfloat *) projectiveShadowMatrix);
							
							drawCone();
							drawSphere();
							drawTori();
							
						glPopMatrix();				
					}
				glDisable(GL_CULL_FACE);
				glDisable(GL_BLEND);
				glEnable(GL_LIGHTING);
			glPopMatrix();
			
			glDisable(GL_STENCIL_TEST);
			
			drawCone();
			drawSphere();
			drawTori();
			DrawRoom();
			
			break;
		
		// depth map shadows - builds the shadow map texture from the depth map,
		// loads the correct texture matrix, draws an ambient light pass, and
		// then using texture coordinate generation and the alpha test, draws
		// a "lit" pass where the shadows are not
		case 1:
			makeShadowMap();

			glMatrixMode(GL_PROJECTION);
			glLoadIdentity();
			gluPerspective(70.0, 1.0, cameraDistance() > 15.0?cameraDistance() - 15.0:0.1, 200.0);

			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			gluLookAt(eye[0], eye[1], eye[2], lookat[0], lookat[1], lookat[2], 0, 1, 0);
			glViewport(viewX, viewY, viewW, viewH);
			
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glBindTexture(GL_TEXTURE_2D, shadowMap);
			
			glMatrixMode(GL_TEXTURE);
			glPushMatrix();
				glLoadMatrixf(textureMatrix);
				glMatrixMode(GL_MODELVIEW);
				glDisable(GL_LIGHT0);
				glEnable(GL_LIGHT1);
				
				glLightfv(GL_LIGHT1, GL_AMBIENT, lowLight);
				glLightfv(GL_LIGHT1, GL_DIFFUSE, lowLight);
				glLightfv(GL_LIGHT1, GL_POSITION, lightPos);
				
				drawCone();
				drawSphere();
				drawTori();
				DrawRoom();
				
				glEnable(GL_BLEND);
				glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_SRC_ALPHA);

				glAlphaFunc(GL_GEQUAL, 0.99);
				glEnable(GL_ALPHA_TEST);
		
				glDisable(GL_LIGHT1);
				glEnable(GL_LIGHT0);
			
				glEnable(GL_TEXTURE_2D);
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			
				glEnable(GL_TEXTURE_GEN_S);
				glEnable(GL_TEXTURE_GEN_T);
				glEnable(GL_TEXTURE_GEN_R);
				glEnable(GL_TEXTURE_GEN_Q);
				glTexGenfv(GL_S, GL_EYE_PLANE, sPlane);
				glTexGenfv(GL_T, GL_EYE_PLANE, tPlane);
				glTexGenfv(GL_R, GL_EYE_PLANE, rPlane);
				glTexGenfv(GL_Q, GL_EYE_PLANE, qPlane);

				drawCone();
				drawSphere();
				drawTori();
				DrawRoom();

				glMatrixMode(GL_TEXTURE);

			glPopMatrix();
			glMatrixMode(GL_MODELVIEW);
			
			glDisable(GL_BLEND);
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_TEXTURE_GEN_S);
			glDisable(GL_TEXTURE_GEN_T);
			glDisable(GL_TEXTURE_GEN_R);
			glDisable(GL_TEXTURE_GEN_Q);
			
			glDisable(GL_ALPHA_TEST);
			break;
		
		// shadow volumes - draws the occluder geometry, and an ambient
		// lighting pass, then creates the shadow volume geometry, performs
		// the zpass shadow volume algorithm, and draws the lit scene where
		// the shadows are not
		case 2:
			drawTriangleOccluder();
			
			glDisable(GL_LIGHT0);
			glEnable(GL_LIGHT1);
			
			glLightfv(GL_LIGHT1, GL_AMBIENT, lowLight);
            glLightfv(GL_LIGHT1, GL_DIFFUSE, lowLight);
			glLightfv(GL_LIGHT1, GL_POSITION, lightPos);
			
			drawCone();
			drawSphere();
			drawTori();
			DrawRoom();


			glDisable(GL_LIGHT1);
			glEnable(GL_LIGHT0);
			
			makeShadowVolume();
		
		if (showVolumeToggle)
		{
			drawShadowVolumeGeometry();			
		}

			break;
	}
}
GLvoid drawMirror(GLvoid)
{
	glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
	glDepthMask(GL_FALSE);

	glClear(GL_STENCIL_BUFFER_BIT);
	glStencilFunc(GL_ALWAYS, 1, 1);
	glStencilOp(GL_REPLACE, GL_ZERO, GL_REPLACE);
	
	glEnable(GL_STENCIL_TEST);
	materialMode(MIRROR_OBJECT);
	glPushMatrix();
		glTranslatef(0.0, 0.0, 19.0);
		glScalef(10.0, 10.0, 0.0);
		glBegin(GL_QUADS);
			glNormal3d(0, 0, 1);			
			glVertex3f( 1.0,-1.0,-1.0);
			glVertex3f(-1.0,-1.0,-1.0);
			glVertex3f(-1.0, 1.0,-1.0);
			glVertex3f( 1.0, 1.0,-1.0);
		glEnd();
	glPopMatrix();
	glClear(GL_DEPTH_BUFFER_BIT);

	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
	glDepthMask(GL_TRUE);

	glStencilFunc(GL_EQUAL, 1, 1);
	glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
	
//	glEnable(GL_BLEND);
//	glBlendFunc(GL_SRC_COLOR, GL_SRC_COLOR);
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glEnable(GL_CULL_FACE);
	glCullFace(GL_FRONT);

	glEnable(GL_LIGHTING);
//   glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(1.0, 0.0);
	glPushMatrix();
		glScalef(1.0, 1.0, -1.0);
	
		glTranslatef(0.0, 0.0, -40.0);

		drawCone();
		drawSphere();
		drawTori();
//		glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	DrawRoom();
	glDisable(GL_LIGHTING);
	glDisable(GL_BLEND);	



	glPopMatrix();
		buildLight(1);
	glDepthMask(GL_TRUE);
	glDisable(GL_POLYGON_OFFSET_FILL);

	glDisable(GL_CULL_FACE);
	glDisable(GL_STENCIL_TEST);
	
//	glutPostRedisplay();
}

/**
*	The display callback for the GLUT display loop.
*	Sets up the eye position, etc, and draws the room
*	with the appropriate shadow method.  A sphere representing
*	the light is added, and the buffers are swapped.
*/
void Display(void)
{
	diffuseLight[0] = diffuseLight[1] = diffuseLight[2] = lightIntensity;
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	
	glMatrixMode(GL_PROJECTION);	
		glLoadIdentity();
		gluPerspective(70.0, 1.0, cameraDistance() > 15.0?cameraDistance() - 15.0:0.1, 200.0);

	glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		gluLookAt(eye[0], eye[1], eye[2], lookat[0], lookat[1], lookat[2], 0, 1, 0);
	
	drawShadows(shadowMode);
//		drawMirror();
	buildLight(0);

	glutSwapBuffers();

}

/**
*	The GLUI callback for GLUI buttons, etc.
*	Resets the camera, and resets the light position.
*
*	@param	control		the GLUI state to control
*/
void glui_cb(int control)
{
	switch(control)
	{
		case 1:
			eye[0] = 0;
			eye[1] = 0;
			eye[2] = -40;
			lookat[0] = 0;
			lookat[1] = -5;
			lookat[2] = 0;
			break;
		case 2:
			lightPositionXY[0] = lightPositionXY[1] = lightPos[0] = lightPos[1] = 10.0;			
			lightPositionZ = lightPos[2] = -10.0;
			
			glutPostRedisplay();
			break;
	}
}


/**
*	The main function.  Sets up the GLUT callback functions,
*	and builds the GLUI GUI interface.
*/
int main(int argc, char** argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH | GLUT_STENCIL);
    glutInitWindowSize(WINDOW_WIDTH + viewX, WINDOW_HEIGHT); 
    glutInitWindowPosition (100, 100);
    
	main_window = glutCreateWindow ("Room with Lights and Tori");
	
	Init();
	
    glutReshapeFunc(myGlutReshape);
	
    glutDisplayFunc(Display); 
	glutMouseFunc(myGlutMouse);
	glutMotionFunc(myGlutMotion);
	
	glui = GLUI_Master.create_glui_subwindow(main_window, GLUI_SUBWINDOW_LEFT);
	glui->add_button("Quit", 0, (GLUI_Update_CB)exit);	
	glui->add_button("Reset Camera", 1, glui_cb);
	
	lightRollout = glui->add_rollout("Light Control");
	lightPanel = glui->add_panel_to_panel(lightRollout, "", GLUI_PANEL_NONE);
	lightSpin = glui->add_spinner_to_panel(lightRollout, "Light Intensity", GLUI_SPINNER_FLOAT, &lightIntensity);
	lightSpin->set_float_limits(0.0, 5.0);
	
	lightTransXY = glui->add_translation_to_panel(lightPanel, "Light XY Position", GLUI_TRANSLATION_XY, lightPositionXY);
	lightTransXY->scale_factor = 0.1;
	
	lightTransZ = glui->add_translation_to_panel(lightPanel, "Light Z Position", GLUI_TRANSLATION_Z, &lightPositionZ);
	lightTransZ->scale_factor = 0.1f;
	glui->add_button_to_panel(lightPanel, "Reset Light", 2, glui_cb);

	shadowRollout = glui->add_rollout("Shadow Control");
	shadowPanel = glui->add_panel_to_panel(shadowRollout, "", GLUI_PANEL_NONE);
	shadowRadio = glui->add_radiogroup_to_panel(shadowPanel, &shadowMode);
	glui->add_radiobutton_to_group(shadowRadio, "Projective Shadows");
	glui->add_radiobutton_to_group(shadowRadio, "Shadow Maps");
	glui->add_radiobutton_to_group(shadowRadio, "Shadow Volumes");

	depthRollout = glui->add_rollout("Extra Features", false);
	depthPanel = glui->add_panel_to_panel(depthRollout, "", GLUI_PANEL_NONE);
	triangleSize = glui->add_spinner_to_panel(depthPanel, "Triangle Size", GLUI_SPINNER_FLOAT, &triangleSizeMultiplier);
	triangleSize->set_float_limits(0.5, 2.5, GLUI_LIMIT_CLAMP);
	volumeCheck = glui->add_checkbox_to_panel(depthPanel, "Show Shadow Volume", &showVolumeToggle);

	glui->add_statictext("");
		
	eye[0] = 0;
	eye[1] = 0;
	eye[2] = -40;
	lookat[0] = 0;
	lookat[1] = -5;
	lookat[2] = 0;

    glutMainLoop();
    
    return 0;
}
